/*-
 *
 * Copyright 2004-2011 Bruce Cran <bruce@cran.org.uk>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <osreldate.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/ata.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <camlib.h>
#include <cam/cam_ccb.h>
#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>

#include "../mi/atagen.h"
#include "ataidle.h"

ATA*
ata_open(const char *path)
{
	ATA *ata = NULL;
	bool using_cam = false;

	ata = malloc(sizeof(ATA));
	if (ata == NULL)
		return NULL;

	char *fullpath = realpath(path, NULL);
	if (fullpath == NULL)
		return NULL;

	char *bp = basename(fullpath);

	/* Check which driver we're using: ata(4), ahci(4) or da(4) */
	if (bp[0] == 'a' && bp[1] == 'd') {
		if (bp[2] == 'a') {
			ata->disk_protocol = DISK_PROTO_ATA_CAM;
			using_cam = true;
		} else {
			ata->disk_protocol = DISK_PROTO_ATA;
			ata->ata_fd = open(fullpath, O_RDONLY);
			if (ata->ata_fd < 0) {
				free(ata);
				free(fullpath);
				return NULL;
			}
		}
	} else {
		ata->disk_protocol = DISK_PROTO_SCSI;
		using_cam = true;
	}

	if (using_cam) {
		ata->cam_dev = cam_open_device( fullpath, O_RDWR );
		if (ata->cam_dev == NULL) {
			free(ata);
			free(fullpath);
			return NULL;
		}
	}

	free(fullpath);
	return ata;
}

void
ata_close( ATA *ata )
{
	if (ata->ata_fd >= 0)
		close(ata->ata_fd);

	free(ata);
}

int
atacmd_ata(ATA *ata, int atacmd, int feature, int feature_val)
{
	int rc;

	ata->atacmd.u.ata.command = atacmd;
	ata->atacmd.u.ata.feature = feature;
	ata->atacmd.u.ata.count = feature_val;

	rc = ioctl(ata->ata_fd, IOCATAREQUEST, &(ata->atacmd));
	return rc;
}

#if __FreeBSD_version > 800000
int
atacmd_cam(ATA *ata, int atacmd, int feature, int feature_val)
{
	union ccb *ccb = cam_getccb(ata->cam_dev);

	int direction;
	int rc = 0;
	uint32_t timeout = ATA_CMD_TIMEOUT*1000;

	if (atacmd == ATA__IDENTIFY || atacmd == ATA__ATAPI_IDENTIFY)
		direction = CAM_DIR_IN;
	else
		direction = CAM_DIR_NONE;

	memset(&(&ccb->ccb_h)[1], 0, sizeof(struct ccb_ataio) - sizeof(struct ccb_hdr));

	cam_fill_ataio(&ccb->ataio,
			5, NULL,
			direction, MSG_SIMPLE_Q_TAG,
			(uint8_t*)ata->outbuffer, ata->outbuffer_size,
			timeout);

	ata_28bit_cmd(&ccb->ataio, atacmd, feature, 0, feature_val);

	ccb->ccb_h.flags |= CAM_DEV_QFRZDIS;

	if (cam_send_ccb(ata->cam_dev, ccb) < 0) {
		cam_error_print(ata->cam_dev, ccb,  CAM_ESF_ALL, CAM_EPF_ALL, stderr);
		rc = 1;
	}

	if (!rc && (ccb->ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_CMP) {
		cam_error_print(ata->cam_dev, ccb, CAM_ESF_ALL, CAM_EPF_ALL, stderr);
		rc = 1;
	}

	cam_freeccb(ccb);

	return rc;
}
#endif

/* send a command to the drive */
int
ata_cmd(ATA *ata, int atacmd, int feature, int feature_val)
{
	int rc = 0;

	if (ata->disk_protocol == DISK_PROTO_ATA)
		rc = atacmd_ata(ata, atacmd, feature, feature_val);
#if __FreeBSD_version > 800000
	else if (ata->disk_protocol == DISK_PROTO_ATA_CAM)
		rc = atacmd_cam(ata, atacmd, feature, feature_val);
#endif

	return rc;

}

/* initialize the ata_cmd structure with supplied values */
void
ata_init(ATA *ata)
{
	if (ata->disk_protocol == DISK_PROTO_ATA) {
		memset(&ata->atacmd, 0, sizeof(struct ata_ioc_request));

		ata->atacmd.u.ata.command = (uint8_t) IOCATAREQUEST;
		ata->atacmd.flags = ATA_CMD_CONTROL;
		ata->atacmd.timeout = ATA_CMD_TIMEOUT;
		ata->atacmd.count = 0;
	}
}


char*
ata_setup_buffer(ATA *ata, int nbytes)
{
	char *databuf = malloc(nbytes);

	if (databuf == NULL)
	{
		/* malloc failed, there's no way to recover here */
		fprintf(stderr, "malloc failed.\n");
		exit(1);
	}

	memset(databuf, 0, nbytes);

	if (ata->disk_protocol == DISK_PROTO_ATA) {
		ata->atacmd.data = databuf;
		ata->atacmd.count = nbytes;
		ata->atacmd.flags = ATA_CMD_READ;
	}

	ata->outbuffer = databuf;
	ata->outbuffer_size = nbytes;

	return databuf;
}

void
ata_free(ATA *ata)
{
	if (ata->outbuffer != NULL)
		free(ata->outbuffer);
}
