/*-
 *
 * Copyright 2004-2010 Bruce Cran <bruce@cran.org.uk>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <arpa/inet.h>
#include <errno.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "atagen.h"

static int is_big_endian(void);
static int ata_getidleval(uint32_t idle_mins, uint16_t *timer_val);
static char* ata_getversionstring(uint16_t ata_version);
static void byteswap(char *buf, int from, int to);
static void strpack(char *buf, int from, int to);
static void byteswap_ata_data(int16_t *buf);
static void mem_swap(int16_t *val);

/* calculate the standby timer value to send to the drive. */
static int
ata_getidleval(uint32_t idle_mins, uint16_t *timer_val)
{
	int rc = 0;

	/* idle value of 0 disables spindown */
	if( idle_mins == 0 )
		*timer_val = 0;

	/* value is (value*5)s between 1 and 20 minutes */
	if( idle_mins <= 20 )
		*timer_val = (idle_mins*60)/5;

	/* special case for 21 minutes */
	if( idle_mins == 21 )
		*timer_val = 252;

	/* there is no encoding for values between 21 and 30 minutes */
	if( (idle_mins > 21) && (idle_mins < 30) ) {
		fprintf(stderr, "Cannot set timeout for 22-29 minutes\n" );
		rc = -1;
	}

	/* after 30 mins, encoding is (idle_mins-29)*30, so you
	 * can only encode multiples of 30 minutes
	 */
	if( idle_mins >= 30 ) {
		/* if it's not a multiple of 30 minutes,
		 * or it's greater than 5 hours, we can't handle it.
		 */
		if ((((idle_mins % 30) != 0) || (idle_mins > 330))
				&& (idle_mins != ATA_IDLEVAL_IMMEDIATE) )
		{
			fprintf(stderr, "Idle timer must be 1-21, or a multiple of 30 "
					"up to 330 (5.5 hours)\n");
			rc = -1;
		}

		/* otherwise, calculate the timer value */
		if(idle_mins == ATA_IDLEVAL_IMMEDIATE)
			*timer_val = ATA_IDLEVAL_IMMEDIATE;
		else
			*timer_val = 240 + (idle_mins/30);
	}

	return rc;
}

static char*
ata_getversionstring(uint16_t ata_version)
{
	const int ATAVERSION_LEN = 16;
	char * version = malloc(ATAVERSION_LEN);
	int i;

	if (version == 0) { /* malloc failed */
		fprintf(stderr, "malloc failed\n");
		return NULL;
	}

	memset(version, 0, ATAVERSION_LEN);

	for (i = 0; i < 15; i++) {
		if ((ata_version >> i) > 0)
			sprintf(version, "ATA-%d", i);
	}

	return version;
}

void
usage(void)
{
	printf("ataidle version " ATAIDLE_VERSION "\n\n"
			"usage: \n"
			"ataidle [-h] [-i] [-s] [-o] [-I standby_mins] [-S standby_mins]\n"
			"        [-A aam] [-P apm] device\n\n"
			"Options:\n");
	printf(
			"-h\t\tdisplay this help\n"
			"-I\t\tenter idle mode and set the standby timer (minutes)\n"
			"-i\t\tenter idle mode\n"
			"-S\t\tenter standby mode and set the standby timer (minutes)\n"
			"-s\t\tenter standby mode\n"
			"-o\t\tenter sleep mode\n"
			"-A\t\tset the AAM (acoustic) level (1-127)\n"
			"-P\t\tset the APM (power) level, (1-254)\n"
			"-W\t\tenable the write cache\n"
			"-w\t\tdisable the write cache\n"
			"device\t\tthe device node e.g. /dev/ada0\n\n"
			"standby_mins is in minutes and has a range of 1-21, then 30-330 in 30\n"
			"minute steps.\n\n"
			"If no options are specified, information about the device will be displayed.\n");

	exit(1);
}

static void
byteswap(char * buf, int from, int to)
{
	int i;

	for (i = from; i <= to; i+=2)
	{
		char b1 = buf[i];
		buf[i] = buf[i+1];
		buf[i+1] = b1;
	}
}

/* Trim leading whitespace */
static void
strpack(char *buf, int from, int to)
{
	int i = 0, j= 0, k = 0;
	int numchars = (to-from)+1;
	char *tmpbuf = malloc(numchars*sizeof(char));

	if(tmpbuf == NULL) { /* malloc failed, skip the string packing */
		fprintf(stderr, "malloc failed\n");
		return;
	}

	strncpy(tmpbuf, buf+from, numchars);
	memset(buf+from, 0, numchars);

	i = 0;

	while(tmpbuf[i] == ' ')
		i++;

	j = 0;
	k = 0;

	for(j = i; j < numchars; j++)
	{
		buf[from+k] = tmpbuf[j];
		k++;
	}

	free(tmpbuf);
}

/* swap the bytes */
static void
mem_swap(int16_t * val)
{
	int8_t * smallval = (int8_t*) val;
	int8_t tmp = smallval[0];

	smallval[0] = smallval[1];
	smallval[1] = tmp;
}

/* detect if we're running on a big-endian platform */
static int
is_big_endian(void)
{
	int i = 0;
	((char*)(&i))[0] = 1;
	return (i != 1);
}

void
ata_showdeviceinfo(ATA *ata, struct ata_ident ident)
{
	int16_t *buf;
	unsigned int mbsize;
	char model[41];
	char serial[21];
	char firmware[9];
	char *ata_version = NULL;

	buf = (int16_t*) &ident;

	memset(model, 0, 41);
	memset(serial, 0, 21);
	memset(firmware, 0, 9);

	strncpy(model, (char*) ident.model, 40);
	strncpy(serial, (char*) ident.serial, 20);
	strncpy(firmware, (char*) ident.firmware, 8);

	printf("Model:\t\t\t%s\n", model);
	printf("Serial:\t\t\t%s\n", serial);
	printf("Firmware Rev:\t\t%s\n", firmware);
	ata_version = ata_getversionstring(ident.version_major);
	printf("ATA revision:\t\t%s\n", (ident.version_major > 1)?
			ata_version : "unknown/pre ATA-2");
	printf("LBA 48:\t\t\t%s\n", (ident.cmd_enabled2 & 0x400)? "yes" : "no");
	printf("Geometry:\t\t%d cyls, %d heads, %d spt\n",
			ident.word1, ident.word3, ident.word6);

	if (ata_version != NULL)
		free(ata_version);

	mbsize = 0;

	if (ident.cmd_enabled2 & 0x400)
	{
		mbsize = (((uint64_t)ident.max_lba48_address[0] +
			((uint64_t)ident.max_lba48_address[1] << 16) +
			((uint64_t)ident.max_lba48_address[2] << 32) +
			((uint64_t)ident.max_lba48_address[3] << 48))*512)/1048576;
	}
	else
	{
		mbsize = ((ident.nsect[0] + (ident.nsect[1] << 16))*512)/1048576;
	}

	bool smart_supported = ident.cmd_supp1 & ATA_SMART_SUPPORTED;
	bool smart_enabled   = ident.cmd_enabled1 & ATA_SMART_ENABLED;
        bool wc_supported    = ident.cmd_supp1 & ATA_WC_SUPPORTED;
        bool wc_enabled      = ident.cmd_enabled1 & ATA_WC_ENABLED;
	bool apm_supported   = ident.cmd_supp2 & ATA_APM_SUPPORTED;
	bool apm_enabled     = ident.cmd_enabled2 & ATA_APM_ENABLED;
	bool aam_supported   = ident.cmd_supp2 & ATA_AAM_SUPPORTED;
	bool aam_enabled     = ident.cmd_enabled2 & ATA_AAM_ENABLED;

	printf("Capacity:\t\t%u%s\n", (mbsize < 1024)?
			mbsize : mbsize/1024, (mbsize < 1024)? "MB" : "GB");

	printf("SMART Supported: \t%s\n", smart_supported? "yes" : "no" );

	if (smart_supported)
		printf("SMART Enabled: \t\t%s\n", smart_enabled? "yes" : "no" );

        printf("Write Cache Supported: \t%s\n", wc_supported? "yes" : "no" );

        if (wc_supported)
                printf("Write Cache Enabled: \t%s\n", wc_enabled? "yes" : "no" );

	printf("APM Supported: \t\t%s\n", apm_supported? "yes" : "no" );

	if (apm_supported) {
		printf("APM Enabled: \t\t%s\n", apm_enabled? "yes" : "no" );
		if (apm_enabled)
			printf("APM Value: \t\t%d\n", ident.apm_value);
	}

	printf("AAM Supported: \t\t%s\n", aam_supported? "yes" : "no" );

	if (aam_supported) {
		printf("AAM Enabled: \t\t%s\n", aam_enabled? "yes" : "no");
		if (aam_enabled)
			printf("AAM Value: \t\t%d\n", ((ident.aam_value & 0x00FF))-127);

		printf("Vendor Recommended AAM: %d\n", ((ident.aam_value & 0xFF00) >> 8)-127);
	}
}

static void
byteswap_ata_data(int16_t * buf)
{
	int i;

	for (i = 0; i < 10; i++) {
		mem_swap(buf+i);
	}

	for (i = 20; i < 23; i++) {
		mem_swap(buf+i);
	}

	for (i = 47; i < 255; i++) {
		mem_swap(buf + i);
	}
}

/*
 * Set the Advanced Power Management mode for the drive.   Modern hard
 * drives can have a number of power management states, ranging from
 * lowest (least performance) to highest power consumption, which results
 * in the highest performance.
 */
int
ata_setapm(ATA *ata, uint32_t apm_val)
{
	int rc = 0;
	int apm_onoff_setting;

	/* user inputs vale 1-126, 0x01-0xFE */

	if (apm_val > ATA_APM_MAXPERF) {
		fprintf(stderr, "Invalid APM value: must be %d-%d\n",
				ATA_APM_MINPERF, ATA_APM_MAXPERF);
		rc = -1;
	}

	/* allocate and initialize the ata_cmd structure */
	ata_init(ata);

	if(apm_val == 0)
		apm_onoff_setting = ATA_APM_DISABLE;
	else
		apm_onoff_setting = ATA_APM_ENABLE;

	/* send the APM command to the drive as a FEATURE */
	if(!rc) {
		rc = ata_cmd(ata, ATA__SETFEATURES, apm_onoff_setting, apm_val);

		if(rc)
			perror("Failed to configure APM");
	}

	return rc;
}

int
ata_setwc(ATA *ata, int enable)
{
	int rc, cmd;

	/* allocate and initialize the ata_cmd structure */
	ata_init(ata);
	cmd = enable ? ATA_WC_ENABLE : ATA_WC_DISABLE;
	rc = ata_cmd(ata, ATA__SETFEATURES, cmd, 0);
	if(rc)
		perror("Failed to configure Write Cache");
	return rc;
}

/*
 * Sets the acoustic level on modern hard drives.   This is used to run it
 * at a lower speed/performance level, which in turn reduces noise.
 */
int
ata_setacoustic(ATA *ata, uint32_t acoustic_val)
{
	int rc = 0;
	int aam_onoff_setting;
	static const int aam_user_offset = 127;

	acoustic_val += aam_user_offset; /* scale it 0x80-0xFE, 128-254 */
	/* range check our acoustic level parameter */
	if (acoustic_val > ATA_AUTOACOUSTIC_MAXPERF) {
		fprintf(stderr, "Invalid acoustic value: must be %d-%d\n",
				1, ATA_AUTOACOUSTIC_MAXPERF - aam_user_offset);
		rc = 1;
	}

	ata_init(ata);

	/* disable AAM if user specified 0, which is 127 once offset */
	if(acoustic_val == 127)
		aam_onoff_setting = ATA_AUTOACOUSTIC_DISABLE;
	else
		aam_onoff_setting = ATA_AUTOACOUSTIC_ENABLE;

	if(!rc) {
		rc = ata_cmd(ata, ATA__SETFEATURES, aam_onoff_setting, acoustic_val);

		if(rc)
			perror("Failed to configure AAM.");
	}
	return rc;
}

/* command the device to spindown after idle_mins of no disk activity */
int
ata_setidletimer(ATA *ata, uint32_t idle_mins)
{
	uint16_t timer_val = 0;
	int rc = 0;

	rc = ata_getidleval( idle_mins, &timer_val );

	if(timer_val == ATA_IDLEVAL_IMMEDIATE)
		ata_init(ata);
	else
		ata_init(ata);

	/* send the IDLE command to the drive */
	if (!rc) {
		if(timer_val == ATA_IDLEVAL_IMMEDIATE)
			rc = ata_cmd(ata, ATA_IDLE_IMMEDIATE, 0, 0);
		else
			rc = ata_cmd(ata, ATA_IDLE, 0, timer_val);
	}

	if (rc)
	{
		fprintf(stderr, "Failed to configure idle mode.\n");
	}

	return rc;
}

/* comand the drive to go into sleep mode */
int
ata_sleep(ATA *ata)
{
	int rc = 0;

	ata_init(ata);

	/* send the SLEEP command to the drive */
	rc = ata_cmd(ata, ATA_SLEEP, 0, 0);

	if (rc)
		perror("Failed to configure sleep mode");

	return rc;
}

/* command the device to spindown after standby_mins of no disk activity */
int
ata_setstandbytimer(ATA *ata, uint32_t standby_mins)
{
	uint16_t timer_val = 0;
	int  rc = 0;

	rc = ata_getidleval( standby_mins, &timer_val );

	if (rc)
		return rc;

	ata_init(ata);

	/* send the STANDBY command to the drive */
	if (!rc)
	{
		if(timer_val == ATA_IDLEVAL_IMMEDIATE)
			rc = ata_cmd(ata, ATA_STANDBY_IMMEDIATE, 0, 0);
		else
			rc = ata_cmd(ata, ATA_STANDBY, 0, timer_val);
	}

	if (rc)
	{
		perror("Failed to configure standby timer");
	}

	return rc;
}

/* this function sends an IDENTIFY command to a drive */
int
ata_identify(ATA *ata, struct ata_ident * identity)
{
	int rc = 0;
	char *buf;

	ata_init(ata);
	buf = ata_setup_buffer(ata, 512);
	rc = ata_cmd(ata, ATA__IDENTIFY, 0, 0);

	if (rc)
	{
		/* then try an ATAPI IDENTIFY, maybe it didn't like
		 * an ATA_IDENTIFY */
		ata_init(ata);
		buf = ata_setup_buffer(ata, 512);
		rc = ata_cmd(ata, ATA__ATAPI_IDENTIFY, 0, 0);
	}

	if (!rc)
	{
		byteswap( buf, 20, 39 ); /* serial */
		byteswap( buf, 46, 52 ); /* firmware */
		byteswap( buf, 54, 92 ); /* model */
		strpack( buf, 20, 39 );
	}

	if (is_big_endian())
		byteswap_ata_data((int16_t*)buf);

	if (!rc)
		memcpy(identity, buf, sizeof(struct ata_ident));

	return rc;
}
